@@ -3,85 +3,73 @@ use warnings;
 
 package Dist::Zilla::Plugin::ReadmeAnyFromPod;
 {
-  $Dist::Zilla::Plugin::ReadmeAnyFromPod::VERSION = '0.131500';
+  $Dist::Zilla::Plugin::ReadmeAnyFromPod::VERSION = '0.133290';
 }
 # ABSTRACT: Automatically convert POD to a README in any format for Dist::Zilla
 
-use Moose;
-use Moose::Autobox;
-use MooseX::Has::Sugar;
-use Moose::Util::TypeConstraints qw(enum);
-use IO::Handle;
 use Encode qw( encode );
+use IO::Handle;
+use List::Util qw( reduce );
+use Moose::Autobox;
+use Moose::Util::TypeConstraints qw(enum);
+use Moose;
+use MooseX::Has::Sugar;
+use PPI::Document;
+use PPI::Token::Pod;
+use Pod::Simple::HTML 3.23;
+use Pod::Simple::Text 3.23;
+use Scalar::Util 'blessed';
 
-# This cannot be the FileGatherer role, because it needs to be called
-# after file munging to get the fully-munged POD.
-with 'Dist::Zilla::Role::InstallTool';
+with 'Dist::Zilla::Role::AfterBuild';
+with 'Dist::Zilla::Role::FileGatherer';
+with 'Dist::Zilla::Role::FileMunger';
 with 'Dist::Zilla::Role::FilePruner';
 
+# TODO: Should these be separate modules?
 our $_types = {
+    pod => {
+        filename => 'README.pod',
+        parser => sub {
+            return $_[0];
+        },
+    },
     text => {
         filename => 'README',
         parser => sub {
-            my $mmcontent = $_[0];
+            my $pod = $_[0];
 
-            require Pod::Text;
-            my $parser = Pod::Text->new();
-            $parser->output_string( \my $input_content );
-            $parser->parse_string_document( $mmcontent );
-
-            my $content;
-            if( defined $parser->{encoding} ){
-                $content = encode( $parser->{encoding} , $input_content );
-            } else {
-                $content = $input_content;
-            }
+            my $parser = Pod::Simple::Text->new;
+            $parser->output_string( \my $content );
+            $parser->parse_characters(1);
+            $parser->parse_string_document($pod);
             return $content;
         },
     },
     markdown => {
         filename => 'README.mkdn',
         parser => sub {
-            my $mmcontent = $_[0];
+            my $pod = $_[0];
 
             require Pod::Markdown;
             my $parser = Pod::Markdown->new();
 
             require IO::Scalar;
-            my $input_handle = IO::Scalar->new(\$mmcontent);
+            my $input_handle = IO::Scalar->new(\$pod);
 
             $parser->parse_from_filehandle($input_handle);
             my $content = $parser->as_markdown();
             return $content;
         },
     },
-    pod => {
-        filename => 'README.pod',
-        parser => sub {
-            my $mmcontent = $_[0];
-
-            require Pod::Select;
-            require IO::Scalar;
-            my $input_handle = IO::Scalar->new(\$mmcontent);
-            my $content = '';
-            my $output_handle = IO::Scalar->new(\$content);
-
-            my $parser = Pod::Select->new();
-            $parser->parse_from_filehandle($input_handle, $output_handle);
-
-            return $content;
-        },
-    },
     html => {
         filename => 'README.html',
         parser => sub {
-            my $mmcontent = $_[0];
+            my $pod = $_[0];
 
-            require Pod::Simple::HTML;
             my $parser = Pod::Simple::HTML->new;
-            my $content;
-            $parser->output_string( \$content );
-            $parser->parse_string_document($mmcontent);
+            $parser->output_string( \my $content );
+            $parser->parse_characters(1);
+            $parser->parse_string_document($pod);
             return $content;
         }
     }
@@ -116,51 +104,110 @@ has location => (
 );
 
 
-sub prune_files {
-  my ($self) = @_;
-  if ($self->location eq 'root') {
-      for my $file ($self->zilla->files->flatten) {
-          next unless $file->name eq $self->filename;
-          $self->log_debug([ 'pruning %s', $file->name ]);
-          $self->zilla->prune_file($file);
-      }
-  }
-  return;
+sub gather_files {
+    my ($self) = @_;
+
+    my $filename = $self->filename;
+    if ( $self->location eq 'build'
+         # allow for the file to also exist in the dist
+         and not @{$self->zilla->files->grep( sub { $_->name eq $filename })}
+       ) {
+        require Dist::Zilla::File::InMemory;
+        my $file = Dist::Zilla::File::InMemory->new({
+            content => 'this will be overwritten',
+            name    => $self->filename,
+        });
+
+        $self->add_file($file);
+    }
+    return;
 }
 
 
-sub setup_installer {
+sub prune_files {
     my ($self) = @_;
 
-    require Dist::Zilla::File::InMemory;
+    # leave the file in the dist if another instance of us is adding it there.
+    if ($self->location eq 'root'
+        and not grep {
+            blessed($self) eq blessed($_)
+                and $_->location eq 'build'
+                and $_->filename eq $self->filename
+        } @{$self->zilla->plugins}) {
+        for my $file ($self->zilla->files->flatten) {
+            next unless $file->name eq $self->filename;
+            $self->log_debug([ 'pruning %s', $file->name ]);
+            $self->zilla->prune_file($file);
+        }
+    }
+    return;
+}
 
-    my $content = $self->get_readme_content();
 
-    my $filename = $self->filename;
-    my $file = $self->zilla->files->grep( sub { $_->name eq $filename } )->head;
+sub munge_files {
+    my $self = shift;
 
     if ( $self->location eq 'build' ) {
-        if ( $file ) {
-            $file->content( $content );
-            $self->log("Override $filename in build");
-        } else {
-            $file = Dist::Zilla::File::InMemory->new({
-                content => $content,
-                name    => $filename,
-            });
-            $self->add_file($file);
-        }
+        my $filename = $self->filename;
+        my $file = $self->zilla->files->grep( sub { $_->name eq $filename } )->head;
+        $self->munge_file($file);
     }
-    elsif ( $self->location eq 'root' ) {
-        require File::Slurp;
+    return;
+}
+
+
+sub munge_file {
+    my ($self, $file) = @_;
+
+    # Ensure that we repeat the munging if the source file is modified
+    # after we run.
+    my $source_file = $self->_source_file();
+    if (not $source_file->does('Dist::Zilla::Role::File::ChangeNotification'))
+    {
+        require Dist::Zilla::Role::File::ChangeNotification;
+        Dist::Zilla::Role::File::ChangeNotification->meta->apply($source_file);
+        my $plugin = $self;
+        $source_file->on_changed(sub {
+            my ($self, $newcontent) = @_;
+
+            # If the new content is actually different, recalculate
+            # the content based on the updates.
+            if ($newcontent ne $plugin->_last_source_content)
+            {
+                $plugin->log('someone tried to munge ' . $source_file->name . ' after we read from it. Making modifications again...');
+                $plugin->munge_file($file);
+            }
+        });
+
+        $source_file->watch_file;
+    }
+
+    $self->log_debug([ 'ReadmeAnyFromPod updating contents of %s in dist', $file->name ]);
+    $file->content($self->get_readme_content);
+    return;
+}
+
+
+sub after_build {
+    my $self = shift;
+
+    if ( $self->location eq 'root' ) {
+        my $filename = $self->filename;
+        $self->log_debug([ 'ReadmeAnyFromPod updating contents of %s in root', $filename ]);
+
+        my $content = $self->get_readme_content();
+
+        require Path::Tiny;
         my $file = $self->zilla->root->file($filename);
         if (-e $file) {
-            $self->log("Override $filename in root");
+            $self->log("overriding $filename in root");
         }
-        File::Slurp::write_file("$file", {binmode => ':raw'}, $content);
-    }
-    else {
-        die "Unknown location specified";
+        my $encoding = $self->_get_source_encoding();
+        Path::Tiny::path($file)->spew_raw(
+            $encoding eq 'raw'
+                ? $content
+                : encode($encoding, $content)
+        );
     }
 
     return;
@@ -171,15 +218,57 @@ sub _file_from_filename {
     for my $file ($self->zilla->files->flatten) {
         return $file if $file->name eq $filename;
     }
-    return; # let moose throw exception if nothing found
+    die 'no README found (place [ReadmeAnyFromPod] below [Readme] in dist.ini)!';
+}
+
+sub _source_file {
+    my ($self) = shift;
+    $self->_file_from_filename($self->source_filename);
+}
+
+# Holds the contents of the source file as of the last time we
+# generated a readme from it. We use this to detect when the source
+# file is modified so we can update the README file again.
+has _last_source_content => (
+    is => 'rw', isa => 'Str',
+    default => '',
+);
+
+sub _get_source_content {
+    my ($self) = shift;
+    $self->_last_source_content($self->_source_file->content);
+}
+
+sub _get_source_pod {
+    my ($self) = shift;
+    my $source_content = $self->_get_source_content();
+
+    my $doc = PPI::Document->new(\$source_content);
+    my $pod_elems = $doc->find('PPI::Token::Pod');
+    my $pod_content = "";
+    if ($pod_elems) {
+        # Concatenation should stringify it
+        $pod_content .= PPI::Token::Pod->merge(@$pod_elems);
+    }
+
+    return $pod_content;
+}
+
+sub _get_source_encoding {
+    my ($self) = shift;
+    return
+        $self->_source_file->can('encoding')
+            ? $self->_source_file->encoding
+            : 'raw';        # Dist::Zilla pre-5.0
 }
 
 
 sub get_readme_content {
     my ($self) = shift;
-    my $mmcontent = $self->_file_from_filename($self->source_filename)->content;
+    my $source_pod = $self->_get_source_pod();
     my $parser = $_types->{$self->type}->{parser};
-    my $readme_content = $parser->($mmcontent);
+    # Save the POD text used to generate the README.
+    return $parser->($source_pod);
 }
 
 {
@@ -217,7 +306,7 @@ Dist::Zilla::Plugin::ReadmeAnyFromPod - Automatically convert POD to a README in
 
 =head1 VERSION
 
-version 0.131500
+version 0.133290
 
 =head1 SYNOPSIS
 
@@ -252,6 +341,14 @@ attributes from it. The format is "Readme[TYPE]In[LOCATION]". The
 words "Readme" and "In" are optional, and the whole name is
 case-insensitive. The SYNOPSIS section above gives one example.
 
+When run with C<location = dist>, this plugin runs in the C<FileMunger> phase
+to create the new file. If it runs before another C<FileMunger> plugin does,
+that happens to modify the input pod (like, say,
+L<C<[PodWeaver]>|Dist::Zilla::Plugin::PodWeaver>), the README file contents
+will be recalculated, along with a warning that you should modify your
+F<dist.ini> by referencing C<[ReadmeAnyFromPod]> lower down in the file (the
+build still works, but is less efficient).
+
 =head1 ATTRIBUTES
 
 =head2 type
@@ -288,15 +385,27 @@ built dist.
 
 =head1 METHODS
 
+=head2 gather_files
+
+We create the file early, so other plugins that need to have the full list of
+files are aware of what we will be generating.
+
 =head2 prune_files
 
 Files with C<location = root> must also be pruned, so that they don't
 sneak into the I<next> build by virtue of already existing in the root
-dir.
+dir.  (The alternative is that the user doesn't add them to the build in the
+first place, with an option to their C<GatherDir> plugin.)
 
-=head2 setup_installer
+=head2 munge_files
 
-Adds the requested README file to the dist.
+=head2 munge_file
+
+Edits the content into the requested README file in the dist.
+
+=head2 after_build
+
+Create the requested README file in the root.
 
 =head2 get_readme_content
 
